"""Matchbox."""

from contextlib import suppress
from importlib.metadata import PackageNotFoundError, version

from matchbox.common.exceptions import MatchboxClientSettingsException
from matchbox.common.logging import logger

try:
    # Environment variables must be loaded first for other imports to work
    from matchbox.client import *  # noqa: E402, F403
except MatchboxClientSettingsException:
    logger.warning(
        "Impossible to initialise client. "
        "Please ignore if running in server mode. Otherwise, check your .env file.",
    )

with suppress(PackageNotFoundError):
    __version__ = version("matchbox-db")



"""Composed PostgreSQL adapter for Matchbox server."""

from matchbox.server.postgresql.adapter.main import (
    MatchboxPostgres,
    MatchboxPostgresSettings,
)

__all__ = ("MatchboxPostgres", "MatchboxPostgresSettings")



"""Resolutions API routes for the Matchbox server."""

import zipfile
from io import BytesIO
from typing import Annotated

from fastapi import (
    APIRouter,
    Query,
    Response,
    status,
)

from matchbox.common.arrow import JudgementsZipFilenames, table_to_buffer
from matchbox.common.dtos import (
    CollectionName,
    ErrorResponse,
    ModelResolutionName,
    ModelResolutionPath,
    RunID,
)
from matchbox.common.eval import Judgement
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    ZipResponse,
)

router = APIRouter(prefix="/eval", tags=["eval"])


@router.post(
    "/judgements",
    responses={404: {"model": ErrorResponse}},
    status_code=status.HTTP_201_CREATED,
)
def insert_judgement(
    backend: BackendDependency,
    judgement: Judgement,
) -> Response:
    """Submit judgement from human evaluator."""
    backend.insert_judgement(judgement=judgement)
    return Response(status_code=status.HTTP_201_CREATED)


@router.get(
    "/judgements",
)
def get_judgements(
    backend: BackendDependency,
    tag: Annotated[
        str | None, Query(description="Tag by which to filter judgements")
    ] = None,
) -> ParquetResponse:
    """Retrieve all judgements from human evaluators."""
    judgements, expansion = backend.get_judgements(tag)
    judgements_buffer, expansion_buffer = (
        table_to_buffer(judgements),
        table_to_buffer(expansion),
    )

    zip_buffer = BytesIO()
    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        zip_file.writestr(JudgementsZipFilenames.JUDGEMENTS, judgements_buffer.read())
        zip_file.writestr(JudgementsZipFilenames.EXPANSION, expansion_buffer.read())

    zip_buffer.seek(0)

    return ZipResponse(zip_buffer.getvalue())


@router.get(
    "/samples",
    responses={404: {"model": ErrorResponse}, 422: {"model": ErrorResponse}},
)
def sample(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ModelResolutionName,
    n: int,
    user_name: str,
) -> ParquetResponse:
    """Sample n cluster to validate."""
    sample = backend.sample_for_eval(
        path=ModelResolutionPath(collection=collection, run=run_id, name=resolution),
        n=n,
        user_name=user_name,
    )
    buffer = table_to_buffer(sample)
    return ParquetResponse(buffer.getvalue())



import polars as pl
import pyarrow as pa
import pytest
from httpx import Response
from pandas import DataFrame as PandasDataFrame
from polars.testing import assert_frame_equal
from respx import MockRouter
from sqlalchemy import Engine
from sqlglot.errors import ParseError

from matchbox.client.queries import Query, QueryCombineType, _clean
from matchbox.common.arrow import (
    SCHEMA_QUERY,
    SCHEMA_QUERY_WITH_LEAVES,
    table_to_buffer,
)
from matchbox.common.dtos import ErrorResponse, QueryConfig
from matchbox.common.exceptions import (
    MatchboxEmptyServerResponse,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.sources import (
    linked_sources_factory,
    source_factory,
    source_from_tuple,
)


def test_init_query() -> None:
    """Test that query is initialised correctly"""
    source = source_factory().source
    model = model_factory(dag=source.dag).model
    query = Query(
        source,
        dag=source.dag,
        model=model,
        combine_type="explode",
        threshold=0.32,
        cleaning={"hello": "hello"},
    )

    assert query.config == QueryConfig(
        source_resolutions=[source.name],
        model_resolution=model.name,
        combine_type="explode",
        threshold=32,
        cleaning={"hello": "hello"},
    )


def test_query_single_source(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Tests that we can query from a single source."""
    # Dummy data and source
    testkit = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_keys=["0", "1"],
        name="foo",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # Mock API
    query_route = matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"key": "0", "id": 1},
                        {"key": "1", "id": 2},
                    ],
                    schema=SCHEMA_QUERY,
                )
            ).read(),
        )
    )
    # Tests with no optional params
    results = testkit.source.query().data()
    assert len(results) == 2
    assert {"foo_a", "foo_b", "id"} == set(results.columns)

    assert dict(query_route.calls.last.request.url.params) == {
        "collection": testkit.source.dag.name,
        "run_id": str(testkit.source.dag.run),
        "source": testkit.source.name,
        "return_leaf_id": "False",
    }

    # Tests with optional params
    results = testkit.source.query(threshold=0.5).data(return_type="pandas")

    assert isinstance(results, PandasDataFrame)
    assert len(results) == 2
    assert {"foo_a", "foo_b", "id"} == set(results.columns)

    assert dict(query_route.calls.last.request.url.params) == {
        "collection": testkit.source.dag.name,
        "run_id": str(testkit.source.dag.run),
        "source": testkit.source.name,
        "threshold": "50",
        "return_leaf_id": "False",
    }


def test_query_multiple_sources(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Tests that we can query multiple sources."""
    # Dummy data and source
    testkit1 = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_keys=["0", "1"],
        name="foo",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    testkit2 = source_from_tuple(
        data_tuple=({"c": "val"}, {"c": "val"}),
        data_keys=["2", "3"],
        name="foo2",
        engine=sqla_sqlite_warehouse,
        dag=testkit1.source.dag,
    ).write_to_location()

    # Mock API
    query_route = matchbox_api.get("/query").mock(
        side_effect=[
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"key": "0", "id": 1},
                            {"key": "1", "id": 2},
                        ],
                        schema=SCHEMA_QUERY,
                    )
                ).read(),
            ),
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"key": "2", "id": 1},
                            {"key": "3", "id": 2},
                        ],
                        schema=SCHEMA_QUERY,
                    )
                ).read(),
            ),
        ]
        * 2  # 2 calls to `query()` in this test, each querying server twice
    )

    model = model_factory(dag=testkit1.source.dag).model
    # Validate results (no cleaning, so all columns passed through)
    results = model.query(testkit1.source, testkit2.source).data()
    assert len(results) == 4
    assert {"foo_a", "foo_b", "foo2_c", "id"} == set(results.columns)

    assert dict(query_route.calls[-2].request.url.params) == {
        "collection": testkit1.source.dag.name,
        "run_id": str(testkit1.source.dag.run),
        "source": testkit1.source.name,
        "resolution": model.name,
        "return_leaf_id": "False",
    }
    assert dict(query_route.calls[-1].request.url.params) == {
        "collection": testkit2.source.dag.name,
        "run_id": str(testkit2.source.dag.run),
        "source": testkit2.source.name,
        "resolution": model.name,
        "return_leaf_id": "False",
    }


def test_query_cleaning(
    sqla_sqlite_warehouse: Engine, matchbox_api: MockRouter
) -> None:
    """Can iterate on cleaning functions."""
    # Set up mocks
    source = (
        source_from_tuple(
            data_tuple=({"col1": " a "}, {"col1": " b "}),
            data_keys=["0", "1"],
            name="foo",
            engine=sqla_sqlite_warehouse,
        )
        .write_to_location()
        .source
    )

    query = source.query(cleaning={"col1": f"upper({source.f('col1')})"})

    matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"key": "0", "id": 1},
                        {"key": "1", "id": 2},
                    ],
                    schema=SCHEMA_QUERY,
                )
            ).read(),
        )
    )

    # Original query can be run
    cleaned1 = query.data()

    cleaned1_expected = pl.DataFrame(
        [
            {"id": 1, "col1": " A "},
            {"id": 2, "col1": " B "},
        ]
    )

    assert_frame_equal(
        cleaned1, cleaned1_expected, check_column_order=False, check_row_order=False
    )

    # After, we can iterate on the cleaning
    query.cleaning = {"col1": f"trim(upper({source.f('col1')}))"}
    cleaned2 = query.data()

    cleaned2_expected = pl.DataFrame(
        [
            {"id": 1, "col1": "A"},
            {"id": 2, "col1": "B"},
        ]
    )

    assert_frame_equal(
        cleaned2, cleaned2_expected, check_column_order=False, check_row_order=False
    )


def test_query_prefetching(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Tests that we can iterate on cleaning without re-fetching raw data."""
    # Dummy data and source
    testkit = source_from_tuple(
        data_tuple=({"a": 1, "b": "2"}, {"a": 10, "b": "20"}),
        data_keys=["0", "1"],
        name="foo",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # Mock API
    query_route = matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist(
                    [
                        {"key": "0", "id": 1},
                        {"key": "1", "id": 2},
                    ],
                    schema=SCHEMA_QUERY,
                )
            ).read(),
        )
    )

    query = testkit.source.query()

    raw = query.data_raw()
    assert query_route.call_count == 1

    data_prefetched = query.data(raw)
    assert query_route.call_count == 1

    assert_frame_equal(query.data(), data_prefetched, check_row_order=False)
    assert query_route.call_count == 2


@pytest.mark.parametrize(
    "combine_type",
    [QueryCombineType.SET_AGG, QueryCombineType.EXPLODE],
    ids=["set_agg", "explode"],
)
def test_query_combine_type(
    combine_type: str, matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Various ways of combining multiple sources are supported."""
    # Dummy data and source
    testkit1 = source_from_tuple(
        data_tuple=({"col": 20}, {"col": 40}, {"col": 60}),
        data_keys=["0", "1", "2"],
        name="foo",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    testkit2 = source_from_tuple(
        data_tuple=({"col": "val1"}, {"col": "val2"}, {"col": "val3"}),
        data_keys=["3", "4", "5"],
        name="bar",
        engine=sqla_sqlite_warehouse,
        dag=testkit1.source.dag,
    ).write_to_location()

    # Mock API
    matchbox_api.get("/query").mock(
        side_effect=[
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"key": "0", "id": 1},
                            {"key": "1", "id": 1},
                            {"key": "2", "id": 2},
                        ],
                        schema=SCHEMA_QUERY,
                    )
                ).read(),
            ),
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            # Creating a duplicate value for the same Matchbox ID
                            {"key": "3", "id": 2},
                            {"key": "3", "id": 2},
                            {"key": "4", "id": 3},
                        ],
                        schema=SCHEMA_QUERY,
                    )
                ).read(),
            ),
        ]  # two sources to query
    )

    model = model_factory(dag=testkit1.source.dag).model

    # Validate results
    results = model.query(
        testkit1.source, testkit2.source, combine_type=combine_type
    ).data()

    if combine_type == "set_agg":
        expected_len = 3

        # Iterate over rows
        for row in results.drop("id").iter_rows(named=True):
            for cell in row.values():
                assert isinstance(cell, list)
                # No duplicates
                assert len(cell) == len(set(cell))

    else:
        expected_len = 5

    assert len(results) == expected_len
    assert {"foo_col", "bar_col", "id"} == set(results.columns)


@pytest.mark.parametrize(
    "combine_type",
    [QueryCombineType.CONCAT, QueryCombineType.SET_AGG, QueryCombineType.EXPLODE],
    ids=["concat", "set_agg", "explode"],
)
def test_query_leaf_ids(
    combine_type: str, matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Leaf IDs can be derived as a query byproduct."""
    testkit1 = source_from_tuple(
        data_tuple=({"col": 20}, {"col": 40}, {"col": 60}),
        data_keys=["0", "1", "2"],
        name="foo",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    testkit2 = source_from_tuple(
        data_tuple=({"col": "val1"}, {"col": "val2"}, {"col": "val3"}),
        data_keys=["3", "4", "5"],
        name="bar",
        engine=sqla_sqlite_warehouse,
        dag=testkit1.source.dag,
    ).write_to_location()

    matchbox_api.get("/query").mock(
        side_effect=[
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            {"key": "0", "id": 12, "leaf_id": 1},
                            {"key": "1", "id": 12, "leaf_id": 2},
                            {"key": "2", "id": 345, "leaf_id": 3},
                        ],
                        schema=SCHEMA_QUERY_WITH_LEAVES,
                    )
                ).read(),
            ),
            Response(
                200,
                content=table_to_buffer(
                    pa.Table.from_pylist(
                        [
                            # Creating a duplicate value for the same Matchbox ID
                            {"key": "3", "id": 345, "leaf_id": 4},
                            {"key": "3", "id": 345, "leaf_id": 5},
                            {"key": "4", "id": 6, "leaf_id": 6},
                        ],
                        schema=SCHEMA_QUERY_WITH_LEAVES,
                    )
                ).read(),
            ),
        ]  # two sources to query
    )

    model = model_factory(dag=testkit1.source.dag).model

    query = model.query(testkit1.source, testkit2.source, combine_type=combine_type)
    data: pl.DataFrame = query.data(return_leaf_id=True)
    assert set(data.columns) == {"foo_col", "bar_col", "id"}

    assert_frame_equal(
        pl.DataFrame(query.leaf_id),
        pl.DataFrame(
            [
                {"leaf_id": 1, "id": 12},
                {"leaf_id": 2, "id": 12},
                {"leaf_id": 3, "id": 345},
                {"leaf_id": 4, "id": 345},
                {"leaf_id": 5, "id": 345},
                {"leaf_id": 6, "id": 6},
            ]
        ),
        check_column_order=False,
        check_row_order=False,
    )


def test_query_404_resolution(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="foo"
    ).write_to_location()

    # Mock API
    matchbox_api.get("/query").mock(
        return_value=Response(
            404,
            json=ErrorResponse(
                exception_type="MatchboxResolutionNotFoundError",
                message="Resolution 42 not found",
            ).model_dump(),
        )
    )

    # Test with no optional params
    with pytest.raises(MatchboxResolutionNotFoundError, match="42"):
        testkit.source.query().data()


def test_query_empty_results_raises_exception(
    matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine
) -> None:
    """Test that query raises MatchboxEmptyServerResponse when no data is returned."""
    testkit = source_factory(
        engine=sqla_sqlite_warehouse, name="foo"
    ).write_to_location()

    # Mock empty results
    matchbox_api.get("/query").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                pa.Table.from_pylist([], schema=SCHEMA_QUERY)
            ).read(),
        )
    )

    # Test that empty results raise MatchboxEmptyServerResponse
    with pytest.raises(
        MatchboxEmptyServerResponse, match="The query operation returned no data"
    ):
        testkit.source.query().data()


def test_query_from_config() -> None:
    """Test reconstructing a Query from a QueryConfig."""
    dag = TestkitDAG().dag

    # Create test sources
    linked_testkit = linked_sources_factory(dag=dag)
    crn_testkit = linked_testkit.sources["crn"]
    dh_testkit = linked_testkit.sources["dh"]

    model_testkit = model_factory(
        left_testkit=crn_testkit,
        right_testkit=dh_testkit,
        true_entities=linked_testkit.true_entities,
        dag=dag,
    )

    # Add to DAG
    dag.source(**crn_testkit.into_dag())
    dag.source(**dh_testkit.into_dag())
    dag.model(**model_testkit.into_dag())

    # Create original query
    original_query = model_testkit.model.query(
        crn_testkit.source,
        dh_testkit.source,
        combine_type="explode",
        threshold=0.75,
        cleaning={"new_col": "foo_a * 2"},
    )

    # Get config and reconstruct
    config = original_query.config
    reconstructed_query = Query.from_config(config, dag=dag)

    # Verify reconstruction matches original
    assert reconstructed_query.config == original_query.config
    assert reconstructed_query.sources == original_query.sources
    assert reconstructed_query.model.config == original_query.model.config
    assert reconstructed_query.combine_type == original_query.combine_type
    assert reconstructed_query.threshold == original_query.threshold
    assert reconstructed_query.cleaning == original_query.cleaning


def test_query_from_config_no_model() -> None:
    """Test reconstructing a Query without a model."""
    dag = TestkitDAG().dag

    # Create test source
    testkit = source_factory(dag=dag)

    # Add to DAG
    dag.source(**testkit.into_dag())

    # Create query without model
    original_query = testkit.source.query(threshold=0.5)

    # Reconstruct from config
    config = original_query.config
    reconstructed_query = Query.from_config(config, dag=dag)

    # Verify
    assert reconstructed_query.config == original_query.config
    assert reconstructed_query.model is None
    assert reconstructed_query.threshold == 0.5


@pytest.mark.parametrize(
    ("cleaning_dict", "expected_columns", "expected_values"),
    [
        pytest.param(
            {"name": "foo_name"},
            ["id", "name"],
            {"name": ["A", "B", "C"]},
            id="simple_column_rename",
        ),
        pytest.param(
            {"upper_name": "upper(foo_name)"},
            ["id", "upper_name"],
            {"upper_name": ["A", "B", "C"]},
            id="simple_transformation",
        ),
        pytest.param(
            {"name": "foo_name", "is_active": "foo_status = 'active'"},
            ["id", "name", "is_active"],
            {"name": ["A", "B", "C"], "is_active": [True, False, True]},
            id="multiple_columns",
        ),
    ],
)
def test_clean_basic_functionality(
    cleaning_dict: dict[str, str],
    expected_columns: list[str],
    expected_values: dict[str, list],
) -> None:
    """Test that clean() basic functionality works."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "foo_name": ["A", "B", "C"],
            "foo_status": ["active", "inactive", "active"],
        }
    )

    result = _clean(test_data, cleaning_dict)
    assert len(result) == 3
    assert set(result.columns) == set(expected_columns)

    for column, values in expected_values.items():
        assert result[column].to_list() == values


def test_clean_none_returns_original() -> None:
    """Test that None cleaning_dict returns original data unchanged."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "name": ["John", "Jane", "Bob"],
            "age": [25, 30, 35],
            "city": ["London", "Hull", "Stratford-upon-Avon"],
        }
    )

    result = _clean(test_data, cleaning_dict=None)

    assert_frame_equal(result, test_data)


def test_clean_column_passthrough() -> None:
    """Test that id is always included plus columns referenced in cleaning_dict."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "name": ["John", "Jane", "Bob"],
            "age": [25, 30, 35],
            "city": ["London", "Hull", "Stratford-upon-Avon"],
        }
    )

    cleaning_dict = {"full_name": "name"}
    result = _clean(test_data, cleaning_dict)

    assert set(result.columns) == {"id", "full_name"}
    assert result["id"].to_list() == [1, 2, 3]
    assert result["full_name"].to_list() == ["John", "Jane", "Bob"]


def test_clean_multiple_column_references() -> None:
    """Test expressions that reference multiple columns."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "first": ["John", "Jane", "Bob"],
            "last": ["Doe", "Smith", "Johnson"],
            "salary": [50000, 60000, 55000],
        }
    )

    cleaning_dict = {
        "name": "first || ' ' || last",
        "high_earner": "salary > 55000",
    }

    result = _clean(test_data, cleaning_dict)

    assert set(result.columns) == {"id", "name", "high_earner"}
    assert result["name"].to_list() == ["John Doe", "Jane Smith", "Bob Johnson"]
    assert result["high_earner"].to_list() == [False, True, False]


def test_clean_complex_sql_expressions() -> None:
    """Test more complex SQL expressions."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "price": [10.5, 20.0, 15.75],
            "quantity": [2, 1, 3],
            "category": ["A", "B", "A"],
        }
    )

    cleaning_dict = {
        "total": "price * quantity",
        "expensive": "price > 15.0",
        "category_upper": "upper(category)",
    }

    result = _clean(test_data, cleaning_dict)

    assert set(result.columns) == {"id", "total", "expensive", "category_upper"}
    assert result["total"].to_list() == [21.0, 20.0, 47.25]
    assert result["expensive"].to_list() == [False, True, True]
    assert result["category_upper"].to_list() == ["A", "B", "A"]


def test_clean_empty_cleaning_dict() -> None:
    """Test with empty cleaning dict returns only id."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "name": ["A", "B", "C"],
            "value": [10, 20, 30],
        }
    )

    result = _clean(test_data, {})

    # Only id is returned when cleaning_dict is empty
    assert set(result.columns) == {"id"}
    assert result["id"].to_list() == [1, 2, 3]


def test_clean_invalid_sql() -> None:
    """Test that invalid SQL raises ParseError."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "name": ["A", "B", "C"],
        }
    )

    cleaning_dict = {
        "invalid": "foo bar baz",  # Invalid SQL
    }

    with pytest.raises(ParseError):
        _clean(test_data, cleaning_dict)


def test_clean_leaf_id_passed_through() -> None:
    """Test that leaf_id is automatically passed through if present."""
    test_data = pl.DataFrame(
        {
            "id": [1, 2, 3],
            "leaf_id": ["a", "b", "c"],
            "value": [10, 20, 30],
            "status": ["active", "inactive", "pending"],
        }
    )

    cleaning_dict = {"processed_value": "value * 2"}

    result = _clean(test_data, cleaning_dict)

    assert set(result.columns) == {"id", "leaf_id", "processed_value"}
    assert result["id"].to_list() == [1, 2, 3]
    assert result["leaf_id"].to_list() == ["a", "b", "c"]
    assert result["processed_value"].to_list() == [20, 40, 60]


def test_clean_multi_source_data() -> None:
    """Test cleaning with data from multiple sources."""
    test_data = pl.DataFrame(
        {
            "id": [1, 1, 2, 2],
            "foo_key": ["a", "a", "b", "b"],
            "bar_key": ["x", "y", "z", "w"],
            "foo_name": ["Alice", "Alice", "Bob", "Bob"],
            "bar_value": [10, 20, 30, 40],
        }
    )

    cleaning_dict = {
        "combined": "foo_name || ': ' || bar_value",
    }

    result = _clean(test_data, cleaning_dict)

    # Id is always included, plus the cleaned column
    assert set(result.columns) == {"id", "combined"}
    assert result["combined"].to_list() == [
        "Alice: 10",
        "Alice: 20",
        "Bob: 30",
        "Bob: 40",
    ]



"""Common Arrow utilities."""

from enum import StrEnum
from io import BytesIO
from typing import Final

import pyarrow as pa
import pyarrow.parquet as pq
from pyarrow import Schema

from matchbox.common.exceptions import MatchboxArrowSchemaMismatch

SCHEMA_QUERY: Final[pa.Schema] = pa.schema(
    [("id", pa.int64()), ("key", pa.large_string())]
)
"""Data transfer schema for root cluster IDs keyed to primary keys."""

SCHEMA_QUERY_WITH_LEAVES = SCHEMA_QUERY.append(pa.field("leaf_id", pa.int64()))
"""Data transfer schema for root cluster IDs keyed to primary keys and leaf IDs."""


SCHEMA_INDEX: Final[pa.Schema] = pa.schema(
    [("hash", pa.large_binary()), ("keys", pa.large_list(pa.large_string()))]
)
"""Data transfer schema for data to be indexed in Matchbox."""

SCHEMA_RESULTS: Final[pa.Schema] = pa.schema(
    [
        ("left_id", pa.uint64()),
        ("right_id", pa.uint64()),
        ("probability", pa.uint8()),
    ]
)
"""Data transfer schema for the results of a deduplication or linking process."""


SCHEMA_JUDGEMENTS: Final[pa.Schema] = pa.schema(
    [
        ("user_name", pa.large_string()),
        ("endorsed", pa.uint64()),
        ("shown", pa.uint64()),
    ]
)
"""Data transfer schema for retrieved evaluation judgements from users."""

SCHEMA_CLUSTER_EXPANSION: Final[pa.Schema] = pa.schema(
    [
        ("root", pa.uint64()),
        ("leaves", pa.list_(pa.uint64())),
    ]
)
"""Data transfer schema for mapping from a cluster ID to all its source cluster IDs."""

SCHEMA_EVAL_SAMPLES: Final[pa.Schema] = pa.schema(
    [
        ("root", pa.uint64()),
        ("leaf", pa.uint64()),
        ("key", pa.large_string()),
        ("source", pa.large_string()),
    ]
)
"""Data transfer schema for evaluation samples."""


class JudgementsZipFilenames(StrEnum):
    """Enumeration of file names in ZIP file with downloaded judgements."""

    JUDGEMENTS = "judgements.parquet"
    EXPANSION = "expansion.parquet"


def table_to_buffer(table: pa.Table) -> BytesIO:
    """Converts an Arrow table to a BytesIO buffer."""
    sink = BytesIO()
    pq.write_table(table, sink)
    sink.seek(0)
    return sink


def check_schema(expected: Schema, actual: Schema) -> None:
    """Validate equality of Arrow schemas, ignoring field order and metadata.

    Args:
        expected: Schema to check against
        actual: Schema to check
    """
    # Make comparison invariant to order
    actual = pa.schema(
        [
            (name, actual.field(name).type)
            for name in expected.names
            if name in actual.names
        ]
    )

    if not actual.equals(expected):
        raise MatchboxArrowSchemaMismatch(expected=expected, actual=actual)



"""Query PostgreSQL mixin for Matchbox server."""

from typing import TYPE_CHECKING, Any

from matchbox.common.dtos import Match, ResolutionPath, SourceResolutionPath
from matchbox.server.postgresql.utils.query import match, query

if TYPE_CHECKING:
    from pyarrow import Table as ArrowTable
else:
    ArrowTable = Any


class MatchboxPostgresQueryMixin:
    """Query mixin for the PostgreSQL adapter for Matchbox."""

    def query(  # noqa: D102
        self,
        source: SourceResolutionPath,
        point_of_truth: ResolutionPath | None = None,
        threshold: int | None = None,
        return_leaf_id: bool = False,
        limit: int | None = None,
    ) -> ArrowTable:
        return query(
            source=source,
            point_of_truth=point_of_truth,
            threshold=threshold,
            return_leaf_id=return_leaf_id,
            limit=limit,
        )

    def match(  # noqa: D102
        self,
        key: str,
        source: SourceResolutionPath,
        targets: list[SourceResolutionPath],
        point_of_truth: ResolutionPath,
        threshold: int | None = None,
    ) -> list[Match]:
        return match(
            key=key,
            source=source,
            targets=targets,
            point_of_truth=point_of_truth,
            threshold=threshold,
        )



"""Test the backend adapter's basic functions."""

from functools import partial

import pytest
from sqlalchemy import Engine
from test.fixtures.db import BACKENDS

from matchbox.common.factories.scenarios import setup_scenario
from matchbox.server.base import MatchboxDBAdapter


@pytest.mark.parametrize("backend", BACKENDS)
@pytest.mark.docker
class TestMatchboxMainBackend:
    @pytest.fixture(autouse=True)
    def setup(
        self, backend_instance: MatchboxDBAdapter, sqla_sqlite_warehouse: Engine
    ) -> None:
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqla_sqlite_warehouse)

    def test_properties(self) -> None:
        """Test that properties obey their protocol restrictions."""
        with self.scenario(self.backend, "index"):
            assert isinstance(self.backend.sources.list_all(), list)
            assert isinstance(self.backend.sources.count(), int)
            assert isinstance(self.backend.models.count(), int)
            assert isinstance(self.backend.source_clusters.count(), int)
            assert isinstance(self.backend.model_clusters.count(), int)
            assert isinstance(self.backend.all_clusters.count(), int)
            assert isinstance(self.backend.creates.count(), int)
            assert isinstance(self.backend.merges.count(), int)
            assert isinstance(self.backend.proposes.count(), int)



"""Renaming sources to source config and source_id to source_config_id.

Revision ID: 95c0b5c23446
Revises: beba75a24962
Create Date: 2025-05-12 12:24:29.340571

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "95c0b5c23446"
down_revision: str | None = "beba75a24962"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # First, drop the foreign key constraints that reference the sources table
    op.drop_constraint(
        "cluster_source_pks_source_id_fkey",
        "cluster_source_pks",
        schema="mb",
        type_="foreignkey",
    )
    op.drop_constraint(
        "source_columns_source_id_fkey",
        "source_columns",
        schema="mb",
        type_="foreignkey",
    )

    # Rename the sources table to source_configs
    op.rename_table("sources", "source_configs", schema="mb")

    # Rename source_id column to source_config_id in all tables
    op.alter_column(
        "source_configs", "source_id", new_column_name="source_config_id", schema="mb"
    )
    op.alter_column(
        "cluster_source_pks",
        "source_id",
        new_column_name="source_config_id",
        schema="mb",
    )
    op.alter_column(
        "source_columns", "source_id", new_column_name="source_config_id", schema="mb"
    )

    # Update indexes to use the new column name
    op.drop_index(
        "ix_source_columns_source_id", table_name="source_columns", schema="mb"
    )
    op.create_index(
        "ix_source_columns_source_config_id",
        "source_columns",
        ["source_config_id"],
        unique=False,
        schema="mb",
    )

    # Update unique constraints to use new column name
    op.drop_constraint(
        "unique_pk_source", "cluster_source_pks", schema="mb", type_="unique"
    )
    op.create_unique_constraint(
        "unique_pk_source",
        "cluster_source_pks",
        ["pk_id", "source_config_id"],
        schema="mb",
    )

    op.drop_constraint(
        "unique_column_index", "source_columns", schema="mb", type_="unique"
    )
    op.create_unique_constraint(
        "unique_column_index",
        "source_columns",
        ["source_config_id", "column_index"],
        schema="mb",
    )

    # Recreate the foreign key constraints with the new column and table names
    op.create_foreign_key(
        "cluster_source_pks_source_id_fkey",
        "cluster_source_pks",
        "source_configs",
        ["source_config_id"],
        ["source_config_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "source_columns_source_id_fkey",
        "source_columns",
        "source_configs",
        ["source_config_id"],
        ["source_config_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )


def downgrade() -> None:
    """Downgrade schema."""
    # Drop the foreign key constraints that reference the source_configs table
    op.drop_constraint(
        "source_columns_source_id_fkey",
        "source_columns",
        schema="mb",
        type_="foreignkey",
    )
    op.drop_constraint(
        "cluster_source_pks_source_id_fkey",
        "cluster_source_pks",
        schema="mb",
        type_="foreignkey",
    )

    # Drop constraints that use the new column names before renaming columns
    op.drop_constraint(
        "unique_column_index", "source_columns", schema="mb", type_="unique"
    )
    op.drop_constraint(
        "unique_pk_source", "cluster_source_pks", schema="mb", type_="unique"
    )
    op.drop_index(
        "ix_source_columns_source_config_id", table_name="source_columns", schema="mb"
    )

    # Rename source_config_id column back to source_id in all tables
    op.alter_column(
        "source_columns", "source_config_id", new_column_name="source_id", schema="mb"
    )
    op.alter_column(
        "cluster_source_pks",
        "source_config_id",
        new_column_name="source_id",
        schema="mb",
    )
    op.alter_column(
        "source_configs", "source_config_id", new_column_name="source_id", schema="mb"
    )

    # Now recreate constraints using the old column names
    op.create_unique_constraint(
        "unique_column_index",
        "source_columns",
        ["source_id", "column_index"],
        schema="mb",
    )
    op.create_unique_constraint(
        "unique_pk_source", "cluster_source_pks", ["pk_id", "source_id"], schema="mb"
    )
    op.create_index(
        "ix_source_columns_source_id",
        "source_columns",
        ["source_id"],
        unique=False,
        schema="mb",
    )

    # Rename the source_configs table back to sources
    op.rename_table("source_configs", "sources", schema="mb")

    # Recreate the original foreign key constraints with the original names
    op.create_foreign_key(
        "source_columns_source_id_fkey",
        "source_columns",
        "sources",
        ["source_id"],
        ["source_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "cluster_source_pks_source_id_fkey",
        "cluster_source_pks",
        "sources",
        ["source_id"],
        ["source_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )



"""Scenario factories for creating TestkitDAG scenarios."""

from collections.abc import Callable, Generator
from contextlib import contextmanager
from typing import Any, Literal

import pyarrow as pa
from polars.testing import assert_frame_equal
from sqlalchemy import Engine

from matchbox.client.queries import Query
from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.entities import (
    FeatureConfig,
    PrefixRule,
    ReplaceRule,
    SuffixRule,
)
from matchbox.common.factories.models import query_to_model_factory
from matchbox.common.factories.sources import (
    SourceTestkitParameters,
    linked_sources_factory,
)
from matchbox.server.base import MatchboxDBAdapter, MatchboxSnapshot

# Type definitions
ScenarioBuilder = Callable[..., TestkitDAG]

SCENARIO_REGISTRY: dict[str, ScenarioBuilder] = {}

# Cache for database snapshots
_DATABASE_SNAPSHOTS_CACHE: dict[str, tuple[TestkitDAG, MatchboxSnapshot]] = {}


def register_scenario(name: str) -> Callable[[ScenarioBuilder], ScenarioBuilder]:
    """Decorator to register a new scenario builder function."""

    def decorator(func: ScenarioBuilder) -> ScenarioBuilder:
        SCENARIO_REGISTRY[name] = func
        return func

    return decorator


def _generate_cache_key(
    backend: MatchboxDBAdapter,
    scenario_type: str,
    warehouse: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> str:
    """Generate a unique hash based on input parameters."""
    if scenario_type not in SCENARIO_REGISTRY:
        raise ValueError(f"Unknown scenario type: {scenario_type}")

    return (
        f"{warehouse.url}_{backend.__class__.__name__}_"
        f"{scenario_type}_{n_entities}_{seed}"
    )


def _testkitdag_to_location(client: Engine, dag: TestkitDAG) -> None:
    """Upload a TestkitDAG to a location warehouse.

    * Writes all data to the location warehouse, replacing existing data
    * Updates the client of all sources in the DAG
    """
    for source_testkit in dag.sources.values():
        source_testkit.write_to_location(set_client=client)


# Scenario builders


@register_scenario("bare")
def create_bare_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a bare TestkitDAG scenario."""
    dag_testkit = TestkitDAG()

    # Create collection and run
    backend.create_collection(name=dag_testkit.dag.name)
    dag_testkit.dag.run = backend.create_run(collection=dag_testkit.dag.name).run_id

    # Create linked sources
    linked = linked_sources_factory(
        n_true_entities=n_entities,
        seed=seed,
        engine=warehouse_engine,
        dag=dag_testkit.dag,
    )
    dag_testkit.add_linked_sources(linked)

    # Write sources to warehouse
    _testkitdag_to_location(warehouse_engine, dag_testkit)

    return dag_testkit


@register_scenario("index")
def create_index_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create an index TestkitDAG scenario."""
    # First create the bare scenario
    dag_testkit = create_bare_scenario(
        backend, warehouse_engine, n_entities, seed, **kwargs
    )

    # Index sources in backend
    for source_testkit in dag_testkit.sources.values():
        backend.create_resolution(
            resolution=source_testkit.fake_run().source.to_resolution(),
            path=source_testkit.resolution_path,
        )
        backend.insert_source_data(
            path=source_testkit.resolution_path,
            data_hashes=source_testkit.data_hashes,
        )

    return dag_testkit


@register_scenario("dedupe")
def create_dedupe_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a dedupe TestkitDAG scenario."""
    # First create the index scenario
    dag_testkit = create_index_scenario(
        backend, warehouse_engine, n_entities, seed, **kwargs
    )

    # Get the LinkedSourcesTestkit using one of the sources
    linked = dag_testkit.source_to_linked["crn"]

    # Create and add deduplication models
    for testkit in dag_testkit.sources.values():
        name = f"naive_test_{testkit.name}"

        # Query the raw data
        source_data = backend.query(source=testkit.resolution_path)

        # Build model testkit using query data
        model_testkit = query_to_model_factory(
            left_query=Query(testkit.source, dag=dag_testkit.dag),
            left_data=source_data,
            left_keys={testkit.name: "key"},
            true_entities=tuple(linked.true_entities),
            name=name,
            description=f"Deduplication of {testkit.name}",
            prob_range=(1.0, 1.0),
            seed=seed,
        )

        # Add to backend and DAG
        backend.create_resolution(
            resolution=model_testkit.fake_run().model.to_resolution(),
            path=model_testkit.resolution_path,
        )
        backend.insert_model_data(
            path=model_testkit.resolution_path,
            results=model_testkit.probabilities.to_arrow(),
        )
        dag_testkit.add_model(model_testkit)

    return dag_testkit


@register_scenario("probabilistic_dedupe")
def create_probabilistic_dedupe_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a probabilistic dedupe TestkitDAG scenario."""
    # First create the index scenario
    dag_testkit = create_index_scenario(
        backend, warehouse_engine, n_entities, seed, **kwargs
    )

    # Get the LinkedSourcesTestkit using one of the sources
    linked = dag_testkit.source_to_linked["crn"]

    # Create and add deduplication models
    for testkit in dag_testkit.sources.values():
        name = f"probabilistic_test_{testkit.name}"

        # Query the raw data
        source_data = backend.query(source=testkit.resolution_path)

        # Build model testkit using query data
        model_testkit = query_to_model_factory(
            left_query=Query(testkit.source, dag=dag_testkit.dag),
            left_data=source_data,
            left_keys={testkit.name: "key"},
            true_entities=tuple(linked.true_entities),
            name=name,
            description=f"Probabilistic deduplication of {testkit.name}",
            prob_range=(0.5, 0.99),
            seed=seed,
        )
        model_testkit.threshold = 50
        assert model_testkit.model.truth == 0.5

        # Add to backend and DAG
        backend.create_resolution(
            resolution=model_testkit.fake_run().model.to_resolution(),
            path=model_testkit.resolution_path,
        )
        backend.insert_model_data(
            path=model_testkit.resolution_path,
            results=model_testkit.probabilities.to_arrow(),
        )
        dag_testkit.add_model(model_testkit)

    return dag_testkit


@register_scenario("link")
def create_link_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a link TestkitDAG scenario."""
    # First create the dedupe scenario
    dag_testkit = create_dedupe_scenario(
        backend, warehouse_engine, n_entities, seed, **kwargs
    )

    # Get the LinkedSourcesTestkit using one of the sources
    linked = dag_testkit.source_to_linked["crn"]

    # Extract models for linking
    crn_model = dag_testkit.models["naive_test_crn"]
    dh_model = dag_testkit.models["naive_test_dh"]
    cdms_model = dag_testkit.models["naive_test_cdms"]

    # Query data for each resolution
    crn_data = backend.query(
        source=dag_testkit.sources["crn"].resolution_path,
        point_of_truth=crn_model.resolution_path,
    )
    dh_data = backend.query(
        source=dag_testkit.sources["dh"].resolution_path,
        point_of_truth=dh_model.resolution_path,
    )
    cdms_data = backend.query(
        source=dag_testkit.sources["cdms"].resolution_path,
        point_of_truth=cdms_model.resolution_path,
    )

    # Create CRN-DH link
    crn_dh_name = "deterministic_naive_test_crn_naive_test_dh"
    crn_dh_model = query_to_model_factory(
        left_query=Query(
            dag_testkit.sources["crn"].source,
            model=crn_model.model,
            dag=dag_testkit.dag,
        ),
        left_data=crn_data,
        left_keys={"crn": "key"},
        right_query=Query(
            dag_testkit.sources["dh"], model=dh_model.model, dag=dag_testkit.dag
        ),
        right_data=dh_data,
        right_keys={"dh": "key"},
        true_entities=tuple(linked.true_entities),
        name=crn_dh_name,
        description="Link between CRN and DH",
        prob_range=(1.0, 1.0),
        seed=seed,
    )

    # Add to backend and DAG
    backend.create_resolution(
        resolution=crn_dh_model.fake_run().model.to_resolution(),
        path=crn_dh_model.resolution_path,
    )
    backend.insert_model_data(
        path=crn_dh_model.resolution_path,
        results=crn_dh_model.probabilities.to_arrow(),
    )
    dag_testkit.add_model(crn_dh_model)

    # Create CRN-CDMS link
    crn_cdms_name = "probabilistic_naive_test_crn_naive_test_cdms"
    crn_cdms_model = query_to_model_factory(
        left_query=Query(
            dag_testkit.sources["crn"].source,
            model=crn_model.model,
            dag=dag_testkit.dag,
        ),
        left_data=crn_data,
        left_keys={"crn": "key"},
        right_query=Query(
            dag_testkit.sources["cdms"].source,
            model=cdms_model.model,
            dag=dag_testkit.dag,
        ),
        right_data=cdms_data,
        right_keys={"cdms": "key"},
        true_entities=tuple(linked.true_entities),
        name=crn_cdms_name,
        description="Link between CRN and CDMS",
        seed=seed,
    )

    # Add to backend and DAG
    crn_cdms_model.threshold = 75
    assert crn_cdms_model.model.truth == 0.75
    backend.create_resolution(
        path=crn_cdms_model.resolution_path,
        resolution=crn_cdms_model.fake_run().model.to_resolution(),
    )
    backend.insert_model_data(
        path=crn_cdms_model.resolution_path,
        results=crn_cdms_model.probabilities.to_arrow(),
    )
    dag_testkit.add_model(crn_cdms_model)

    # Create final join
    # Query the previous link's results
    crn_cdms_data_crn_only = backend.query(
        source=dag_testkit.sources["crn"].resolution_path,
        point_of_truth=crn_cdms_model.resolution_path,
    ).rename_columns(["id", "keys_crn"])
    crn_cdms_data_cdms_only = backend.query(
        source=dag_testkit.sources["cdms"].resolution_path,
        point_of_truth=crn_cdms_model.resolution_path,
    ).rename_columns(["id", "keys_cdms"])
    crn_cdms_data = pa.concat_tables(
        [crn_cdms_data_crn_only, crn_cdms_data_cdms_only],
        promote_options="default",
    ).combine_chunks()

    dh_data_linked = backend.query(
        source=dag_testkit.sources["dh"].resolution_path,
        point_of_truth=dh_model.resolution_path,
    )

    final_join_name = "final_join"
    final_join_model = query_to_model_factory(
        left_query=Query(
            dag_testkit.sources["crn"].source,
            dag_testkit.sources["cdms"].source,
            model=crn_cdms_model.model,
            dag=dag_testkit.dag,
        ),
        left_data=crn_cdms_data,
        left_keys={"crn": "keys_crn", "cdms": "keys_cdms"},
        right_query=Query(
            dag_testkit.sources["dh"].source,
            model=dh_model.model,
            dag=dag_testkit.dag,
        ),
        right_data=dh_data_linked,
        right_keys={"dh": "key"},
        true_entities=tuple(linked.true_entities),
        name=final_join_name,
        description="Final join of all entities",
        seed=seed,
    )

    # Add to backend and DAG
    backend.create_resolution(
        resolution=final_join_model.fake_run().model.to_resolution(),
        path=final_join_model.resolution_path,
    )
    backend.insert_model_data(
        path=final_join_model.resolution_path,
        results=final_join_model.probabilities.to_arrow(),
    )
    dag_testkit.add_model(final_join_model)

    return dag_testkit


@register_scenario("alt_dedupe")
def create_alt_dedupe_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a TestkitDAG scenario with two alternative dedupers."""
    dag_testkit = TestkitDAG()

    # Create collection and run
    backend.create_collection(name=dag_testkit.dag.name)
    dag_testkit.dag.run = backend.create_run(collection=dag_testkit.dag.name).run_id

    # Create linked sources
    company_name_feature = FeatureConfig(
        name="company_name", base_generator="company"
    ).add_variations(SuffixRule(suffix=" UK"))

    foo_a_tkit_source = SourceTestkitParameters(
        name="foo_a",
        engine=warehouse_engine,
        features=(company_name_feature,),
        drop_base=False,
        n_true_entities=n_entities,
        repetition=1,
    )

    linked = linked_sources_factory(
        source_parameters=(foo_a_tkit_source,), dag=dag_testkit.dag
    )
    dag_testkit.add_linked_sources(linked)

    # Write sources to warehouse
    _testkitdag_to_location(warehouse_engine, dag_testkit)

    # Index sources in backend
    for source_testkit in dag_testkit.sources.values():
        backend.create_resolution(
            resolution=source_testkit.fake_run().source.to_resolution(),
            path=source_testkit.resolution_path,
        )
        backend.insert_source_data(
            path=source_testkit.resolution_path,
            data_hashes=source_testkit.data_hashes,
        )

    # Create and add deduplication models
    for testkit in dag_testkit.sources.values():
        model_name1 = f"dedupe_{testkit.name}"
        model_name2 = f"dedupe2_{testkit.name}"

        # Query the raw data
        source_data = backend.query(source=testkit.resolution_path)

        # Build model testkit using query data
        model_testkit1 = query_to_model_factory(
            left_query=Query(testkit.source, dag=dag_testkit.dag),
            left_data=source_data,
            left_keys={testkit.name: "key"},
            true_entities=tuple(linked.true_entities),
            name=model_name1,
            description=f"Deduplication of {testkit.name}",
            prob_range=(0.5, 1.0),
            seed=seed,
        )

        model_testkit2 = query_to_model_factory(
            left_query=Query(testkit.source, dag=dag_testkit.dag),
            left_data=source_data,
            left_keys={testkit.name: "key"},
            true_entities=tuple(linked.true_entities),
            name=model_name2,
            description=f"Deduplication of {testkit.name}",
            prob_range=(0.5, 1.0),
            seed=seed,
        )

        assert len(model_testkit1.probabilities) > 0
        assert_frame_equal(model_testkit1.probabilities, model_testkit2.probabilities)

        for model, threshold in ((model_testkit1, 50), (model_testkit2, 75)):
            model.threshold = threshold
            assert model.model.truth == threshold / 100

            # Add both models to backend and DAG
            backend.create_resolution(
                path=model.resolution_path,
                resolution=model.fake_run().model.to_resolution(),
            )
            backend.insert_model_data(
                path=model.resolution_path, results=model.probabilities.to_arrow()
            )

            # Add to DAG
            dag_testkit.add_model(model)

    return dag_testkit


@register_scenario("convergent_partial")
def create_convergent_partial_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a TestkitDAG scenario with convergent sources.

    This is where two sources index almost identically. TestkitDAG contains two
    indexed sources with repetition, and two naive dedupe models that haven't yet
    had their results inserted.
    """
    dag_testkit = TestkitDAG()

    # Create collection and run
    backend.create_collection(name=dag_testkit.dag.name)
    dag_testkit.dag.run = backend.create_run(collection=dag_testkit.dag.name).run_id

    # Create linked sources
    company_name_feature = FeatureConfig(
        name="company_name", base_generator="company"
    ).add_variations(SuffixRule(suffix=" UK"))

    foo_a_tkit_source = SourceTestkitParameters(
        name="foo_a",
        engine=warehouse_engine,
        features=(company_name_feature,),
        drop_base=False,
        n_true_entities=n_entities,
        repetition=1,
    )

    linked = linked_sources_factory(
        source_parameters=(
            foo_a_tkit_source,
            foo_a_tkit_source.model_copy(update={"name": "foo_b"}),
        ),
        dag=dag_testkit.dag,
    )

    dag_testkit.add_linked_sources(linked)

    # Write sources to warehouse
    _testkitdag_to_location(warehouse_engine, dag_testkit)

    # Index sources in backend
    for source_testkit in dag_testkit.sources.values():
        backend.create_resolution(
            resolution=source_testkit.fake_run().source.to_resolution(),
            path=source_testkit.resolution_path,
        )
        backend.insert_source_data(
            path=source_testkit.resolution_path,
            data_hashes=source_testkit.data_hashes,
        )

        # Query the raw data
        source_query = backend.query(source=source_testkit.resolution_path)

        # Build model testkit using query data
        model_testkit = query_to_model_factory(
            left_query=Query(source_testkit.source, dag=dag_testkit.dag),
            left_data=source_query,
            left_keys={source_testkit.name: "key"},
            true_entities=tuple(linked.true_entities),
            name=f"naive_test_{source_testkit.name}",
            description=f"Deduplication of {source_testkit.name}",
            prob_range=(1.0, 1.0),
            seed=seed,
        )

        assert len(model_testkit.probabilities) > 0

        # Add to DAG
        dag_testkit.add_model(model_testkit)

    return dag_testkit


@register_scenario("convergent")
def create_convergent_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a TestkitDAG scenario with convergent sources, deduped.

    This is where two sources index almost identically. TestkitDAG contains two
    indexed sources with repetition, and two naive dedupe models, all inserted.
    """
    dag_testkit = create_convergent_partial_scenario(
        backend, warehouse_engine, n_entities, seed, **kwargs
    )
    for model_testkit in dag_testkit.models.values():
        backend.create_resolution(
            resolution=model_testkit.fake_run().model.to_resolution(),
            path=model_testkit.resolution_path,
        )
        backend.insert_model_data(
            path=model_testkit.resolution_path,
            results=model_testkit.probabilities.to_arrow(),
        )

    return dag_testkit


@register_scenario("mega")
def create_mega_scenario(
    backend: MatchboxDBAdapter,
    warehouse_engine: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: Any,
) -> TestkitDAG:
    """Create a TestkitDAG scenario that produces large clusters.

    Aims to produce "mega" clusters with more features than the CLI has screen rows,
    and more variations than the CLI has screen columns.

    Args:
        backend: MatchboxDBAdapter instance
        warehouse_engine: SQLAlchemy engine for data warehouse
        n_entities: Number of true product entities to generate
        seed: Random seed for reproducibility
        **kwargs: Additional arguments

    Returns:
        TestkitDAG with extensive product data and linking model
    """
    dag_testkit = TestkitDAG()

    # Create collection and run
    backend.create_collection(name=dag_testkit.dag.name)
    dag_testkit.dag.run = backend.create_run(collection=dag_testkit.dag.name).run_id

    # ===== FEATURES WITH VARIATIONS (4 string features, 1 variation each) =====

    product_name = FeatureConfig(
        name="product_name",
        base_generator="text",
        parameters=(("max_nb_chars", 20),),
    ).add_variations(
        SuffixRule(suffix=" Pro"),
    )

    product_sku = FeatureConfig(
        name="product_sku",
        base_generator="bothify",
        parameters=(("text", "SKU-####-???"),),
    ).add_variations(
        ReplaceRule(old="-", new=""),
    )

    primary_colour = FeatureConfig(
        name="primary_colour",
        base_generator="color_name",
    ).add_variations(
        PrefixRule(prefix="Colour: "),
    )

    primary_material = FeatureConfig(
        name="primary_material",
        base_generator="random_element",
        parameters=(
            (
                "elements",
                (
                    "Plastic",
                    "Metal",
                    "Wood",
                    "Glass",
                ),
            ),
        ),
        unique=False,
    ).add_variations(
        PrefixRule(prefix="Material: "),
    )

    # ===== FEATURES WITHOUT VARIATIONS (46 features) =====

    # Numeric features
    def _numeric_feature(name: str, pattern: str, **kwargs: Any) -> FeatureConfig:
        """Helper to create a numeric feature with bothify pattern."""
        return FeatureConfig(
            name=name,
            base_generator="bothify",
            parameters=(("text", pattern),),
            **kwargs,
        )

    height_cm = FeatureConfig(
        name="height_cm",
        base_generator="pyfloat",
        parameters=(
            ("min_value", 1.0),
            ("max_value", 500.0),
            ("right_digits", 1),
        ),
    )
    manufacturer_code = _numeric_feature("manufacturer_code", "MFR-###??")
    width_cm = _numeric_feature("width_cm", "###.#")
    depth_cm = _numeric_feature("depth_cm", "###.#")
    weight_kg = _numeric_feature("weight_kg", "##.##")
    capacity_litres = _numeric_feature("capacity_litres", "##.#")
    wattage = _numeric_feature("wattage", "####")
    max_temp_c = _numeric_feature("max_temp_c", "###")
    battery_life_hours = _numeric_feature("battery_life_hours", "###")
    quality_rating = _numeric_feature("quality_rating", "#.#")
    speed_rpm = _numeric_feature("speed_rpm", "#####")
    pressure_psi = _numeric_feature("pressure_psi", "###")
    decibel_level = _numeric_feature("decibel_level", "###")
    gauge = _numeric_feature("gauge", "##")

    # Simple text features
    secondary_colour = FeatureConfig(
        name="secondary_colour", base_generator="color_name"
    )
    origin_country = FeatureConfig(name="origin_country", base_generator="country")
    brand_name = FeatureConfig(name="brand_name", base_generator="company")

    # Categorical features
    def _categorical_feature(name: str, *choices: str, **kwargs: Any) -> FeatureConfig:
        """Helper to create a categorical feature with random_element."""
        return FeatureConfig(
            name=name,
            base_generator="random_element",
            parameters=(("elements", choices),),
            unique=False,
            **kwargs,
        )

    voltage = _categorical_feature("voltage", "110V", "220V", "240V", "12V", "24V")
    warranty_months = _categorical_feature("warranty_months", "12", "24", "36")
    shipping_days = _categorical_feature("shipping_days", "1", "2", "3")
    box_size = _categorical_feature("box_size", "Small", "Medium", "Large")
    safety_cert = _categorical_feature("safety_cert", "CE", "UL", "FCC")
    frequency_hz = _categorical_feature("frequency_hz", "50", "60")
    thread_count = _categorical_feature("thread_count", "100", "200", "300")
    ply_count = _categorical_feature("ply_count", "1", "2", "3")
    hardness_rating = _categorical_feature("hardness_rating", "Soft", "Medium", "Hard")
    water_resistance = _categorical_feature("water_resistance", "IPX4", "IPX5")
    dust_rating = _categorical_feature("dust_rating", "IP5X", "IP6X")
    flex_rating = _categorical_feature("flex_rating", "Rigid", "Flexible", "Ultra-flex")
    texture = _categorical_feature("texture", "Matte", "Glossy", "Satin")
    transparency = _categorical_feature("transparency", "Opaque", "Translucent")
    reflectivity = _categorical_feature("reflectivity", "Non-reflective", "Reflective")
    elasticity = _categorical_feature("elasticity", "Rigid", "Elastic")
    conductivity = _categorical_feature("conductivity", "Conductive", "Insulating")
    magnetic_property = _categorical_feature("magnetic_property", "Magnetic")
    thermal_conductivity = _categorical_feature("thermal_conductivity", "High", "Low")
    uv_resistance = _categorical_feature("uv_resistance", "Standard", "Indoor-only")
    chemical_resistance = _categorical_feature("chemical_resistance", "High", "Low")
    recyclability = _categorical_feature("recyclability", "Recyclable")
    fire_rating = _categorical_feature("fire_rating", "Class A", "Class B")
    allergen_info = _categorical_feature("allergen_info", "Hypoallergenic", "Standard")
    assembly_required = _categorical_feature("assembly_required", "Full assembly")
    tool_requirements = _categorical_feature("tool_requirements", "Basic tools")
    skill_level = _categorical_feature("skill_level", "Beginner", "Advanced", "Expert")
    age_range = _categorical_feature("age_range", "0-3", "3-6", "Adult", "All ages")
    user_capacity = _categorical_feature("user_capacity", "1", "2", "3")

    # ===== CREATE TWO SOURCES WITH OVERLAPPING FEATURES =====

    # Source A: Marketplace A
    marketplace_a_params = SourceTestkitParameters(
        name="marketplace_a",
        engine=warehouse_engine,
        features=(
            product_name,
            product_sku,
            manufacturer_code,
            height_cm,
            width_cm,
            depth_cm,
            weight_kg,
            primary_colour,
            secondary_colour,
            primary_material,
            capacity_litres,
            wattage,
            voltage,
            max_temp_c,
            battery_life_hours,
            warranty_months,
            shipping_days,
            box_size,
            quality_rating,
            safety_cert,
            origin_country,
            brand_name,
            speed_rpm,
            pressure_psi,
            frequency_hz,
            decibel_level,
            thread_count,
        ),
        n_true_entities=n_entities,
        repetition=0,
        drop_base=False,
    )

    # Source B: Marketplace B
    marketplace_b_params = SourceTestkitParameters(
        name="marketplace_b",
        engine=warehouse_engine,
        features=(
            product_name,
            product_sku,
            manufacturer_code,
            height_cm,
            width_cm,
            weight_kg,
            primary_colour,
            primary_material,
            ply_count,
            gauge,
            hardness_rating,
            water_resistance,
            dust_rating,
            flex_rating,
            texture,
            transparency,
            reflectivity,
            elasticity,
            conductivity,
            magnetic_property,
            thermal_conductivity,
            uv_resistance,
            chemical_resistance,
            recyclability,
            fire_rating,
            allergen_info,
            assembly_required,
            tool_requirements,
            skill_level,
            age_range,
            user_capacity,
        ),
        n_true_entities=n_entities,
        repetition=0,
        drop_base=False,
    )

    # Create linked sources
    linked = linked_sources_factory(
        source_parameters=(marketplace_a_params, marketplace_b_params),
        dag=dag_testkit.dag,
        seed=seed,
    )
    dag_testkit.add_linked_sources(linked)

    # Write sources to warehouse
    for source_testkit in dag_testkit.sources.values():
        source_testkit.write_to_location(set_client=warehouse_engine)

    # Index sources in backend
    for source_testkit in dag_testkit.sources.values():
        backend.create_resolution(
            resolution=source_testkit.fake_run().source.to_resolution(),
            path=source_testkit.resolution_path,
        )
        backend.insert_source_data(
            path=source_testkit.resolution_path,
            data_hashes=source_testkit.data_hashes,
        )

    # ===== CREATE LINKING MODEL =====

    # Get sources for linking
    marketplace_a = dag_testkit.sources["marketplace_a"]
    marketplace_b = dag_testkit.sources["marketplace_b"]

    # Query both sources
    marketplace_a_data = backend.query(source=marketplace_a.resolution_path)
    marketplace_b_data = backend.query(source=marketplace_b.resolution_path)

    # Create linking model
    link_model = query_to_model_factory(
        left_query=Query(marketplace_a.source, dag=dag_testkit.dag),
        left_data=marketplace_a_data,
        left_keys={"marketplace_a": "key"},
        right_query=Query(marketplace_b.source, dag=dag_testkit.dag),
        right_data=marketplace_b_data,
        right_keys={"marketplace_b": "key"},
        true_entities=tuple(linked.true_entities),
        name="mega_product_linker",
        description="Links products across marketplace_a and marketplace_b catalogues",
        prob_range=(1.0, 1.0),
        seed=seed,
    )

    # Add model to backend
    backend.create_resolution(
        resolution=link_model.fake_run().model.to_resolution(),
        path=link_model.resolution_path,
    )
    backend.insert_model_data(
        path=link_model.resolution_path,
        results=link_model.probabilities.to_arrow(),
    )
    dag_testkit.add_model(link_model)

    return dag_testkit


@contextmanager
def setup_scenario(
    backend: MatchboxDBAdapter,
    scenario_type: Literal[
        "bare",
        "index",
        "dedupe",
        "link",
        "probabilistic_dedupe",
        "alt_dedupe",
        "convergent",
    ],
    warehouse: Engine,
    n_entities: int = 10,
    seed: int = 42,
    **kwargs: dict[str, Any],
) -> Generator[TestkitDAG, None, None]:
    """Context manager for creating TestkitDAG scenarios."""
    if scenario_type not in SCENARIO_REGISTRY:
        raise ValueError(f"Unknown scenario type: {scenario_type}")

    # Generate cache key for backend snapshot
    cache_key = _generate_cache_key(
        backend, scenario_type, warehouse, n_entities, seed, **kwargs
    )

    # Check if we have a backend snapshot cached
    if cache_key in _DATABASE_SNAPSHOTS_CACHE:
        # Load cached snapshot and DAG
        dag_testkit, snapshot = _DATABASE_SNAPSHOTS_CACHE[cache_key]
        # Because the location client can't be deep-copied, this will reuse an
        # old engine, with the same URL as `warehouse`, but a different object
        # Thus, in _testkitdag_to_location we need to overwrite all testkits with
        # our new warehouse object
        dag_testkit = dag_testkit.model_copy(deep=True)

        # Restore backend and write sources to warehouse
        backend.restore(snapshot=snapshot)
        _testkitdag_to_location(warehouse, dag_testkit)
    else:
        # Create new TestkitDAG with proper backend integration
        scenario_builder = SCENARIO_REGISTRY[scenario_type]
        dag_testkit = scenario_builder(backend, warehouse, n_entities, seed, **kwargs)

        # Cache the snapshot and DAG
        _DATABASE_SNAPSHOTS_CACHE[cache_key] = (dag_testkit, backend.dump())

    try:
        yield dag_testkit
    finally:
        backend.clear(certain=True)



"""Base classes and utilities for Matchbox database adapters."""

import json
from abc import ABC, abstractmethod
from enum import StrEnum
from typing import TYPE_CHECKING, Any, Literal, Protocol, Self

import boto3
from botocore.exceptions import ClientError
from pyarrow import Table
from pydantic import BaseModel, Field, SecretStr, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from matchbox.common.dtos import (
    BackendResourceType,
    Collection,
    CollectionName,
    Group,
    GroupName,
    Match,
    ModelResolutionPath,
    PermissionGrant,
    PermissionType,
    Resolution,
    ResolutionPath,
    Run,
    RunID,
    SourceResolutionPath,
    UploadStage,
    User,
)
from matchbox.common.eval import Judgement
from matchbox.common.logging import LogLevelType

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any


PERMISSION_GRANTS: dict[PermissionType, list[PermissionType]] = {
    PermissionType.READ: [
        PermissionType.READ,
        PermissionType.WRITE,
        PermissionType.ADMIN,
    ],
    PermissionType.WRITE: [PermissionType.WRITE, PermissionType.ADMIN],
    PermissionType.ADMIN: [PermissionType.ADMIN],
}
"""A global variable that defines the permission hierarchy.

Keys are the permission, values are a list of permissions that would
grant the permission.

For example, only `PermissionType.ADMIN` can grant `PermissionType.ADMIN`,
but any permission implies `PermissionType.READ`.
"""


class MatchboxBackends(StrEnum):
    """The available backends for Matchbox."""

    POSTGRES = "postgres"


class MatchboxSnapshot(BaseModel):
    """A snapshot of the Matchbox database."""

    backend_type: MatchboxBackends
    data: dict[str, Any]

    @field_validator("data")
    @classmethod
    def check_serialisable(cls, value: dict[str, Any]) -> dict[str, Any]:
        """Validate that the value can be serialised to JSON."""
        try:
            json.dumps(value)
            return value
        except (TypeError, OverflowError) as e:
            raise ValueError(f"Value is not JSON serialisable: {e}") from e


class MatchboxDatastoreSettings(BaseSettings):
    """Settings specific to the datastore configuration."""

    host: str | None = None
    port: int | None = None
    access_key_id: SecretStr | None = None
    secret_access_key: SecretStr | None = None
    default_region: str | None = None
    cache_bucket_name: str

    def get_client(self) -> S3Client:
        """Returns an S3 client for the datastore.

        Creates S3 buckets if they don't exist.
        """
        kwargs = {
            "endpoint_url": f"http://{self.host}:{self.port}"
            if self.host and self.port
            else None,
            "aws_access_key_id": self.access_key_id.get_secret_value()
            if self.access_key_id
            else None,
            "aws_secret_access_key": self.secret_access_key.get_secret_value()
            if self.secret_access_key
            else None,
            "region_name": self.default_region,
        }

        client: S3Client = boto3.client("s3", **kwargs)

        try:
            client.head_bucket(Bucket=self.cache_bucket_name)
        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":
                client.create_bucket(
                    Bucket=self.cache_bucket_name,
                    CreateBucketConfiguration={
                        "LocationConstraint": self.default_region
                    },
                )
            else:
                raise e

        return client


class MatchboxServerSettings(BaseSettings):
    """Settings for the Matchbox application."""

    model_config = SettingsConfigDict(
        env_prefix="MB__SERVER__",
        env_nested_delimiter="__",
        use_enum_values=True,
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    batch_size: int = Field(default=250_000)
    backend_type: MatchboxBackends
    datastore: MatchboxDatastoreSettings
    task_runner: Literal["api", "celery"]
    redis_uri: str | None
    uploads_expiry_minutes: int | None
    authorisation: bool = True
    public_key: SecretStr | None = Field(default=None)
    log_level: LogLevelType = "INFO"

    @model_validator(mode="after")
    def check_settings(self) -> Self:
        """Check that legal combinations of settings are provided."""
        if self.task_runner == "celery" and self.redis_uri is None:
            raise ValueError("A Redis URI must be set if using Celery as task runner.")
        if self.task_runner == "celery" and self.uploads_expiry_minutes is None:
            raise ValueError(
                "Upload expiration must be set if using Celery as task runner."
            )

        return self


class BackendManager:
    """Manages the Matchbox backend instance and settings."""

    _instance = None
    _settings = None

    @classmethod
    def initialise(cls, settings: "MatchboxServerSettings") -> None:
        """Initialise the backend with the given settings."""
        cls._settings = settings

    @classmethod
    def get_backend(cls) -> "MatchboxDBAdapter":
        """Get the backend instance."""
        if cls._settings is None:
            raise ValueError("BackendManager must be initialized with settings first")

        if cls._instance is None:
            BackendClass = get_backend_class(cls._settings.backend_type)
            cls._instance = BackendClass(cls._settings)
        return cls._instance

    @classmethod
    def get_settings(cls) -> "MatchboxServerSettings":
        """Get the backend settings."""
        if cls._settings is None:
            raise ValueError("BackendManager must be initialized with settings first")
        return cls._settings


def get_backend_settings(
    backend_type: MatchboxBackends,
) -> type[MatchboxServerSettings]:
    """Get the appropriate settings class based on the backend type."""
    if backend_type == MatchboxBackends.POSTGRES:
        from matchbox.server.postgresql import MatchboxPostgresSettings  # noqa: PLC0415

        return MatchboxPostgresSettings
    # Add more backend types here as needed
    else:
        raise ValueError(f"Unsupported backend type: {backend_type}")


def get_backend_class(backend_type: MatchboxBackends) -> type["MatchboxDBAdapter"]:
    """Get the appropriate backend class based on the backend type."""
    if backend_type == MatchboxBackends.POSTGRES:
        from matchbox.server.postgresql import MatchboxPostgres  # noqa: PLC0415

        return MatchboxPostgres
    # Add more backend types here as needed
    else:
        raise ValueError(f"Unsupported backend type: {backend_type}")


def settings_to_backend(settings: MatchboxServerSettings) -> "MatchboxDBAdapter":
    """Create backend adapter with injected settings."""
    BackendClass = get_backend_class(settings.backend_type)
    return BackendClass(settings)


def initialise_matchbox() -> None:
    """Initialise the Matchbox backend based on environment variables."""
    base_settings = MatchboxServerSettings()

    SettingsClass = get_backend_settings(base_settings.backend_type)
    settings = SettingsClass()

    BackendManager.initialise(settings)


class Countable(Protocol):
    """A protocol for objects that can be counted."""

    def count(self) -> int:
        """Counts the number of items in the object."""
        ...


class Listable(Protocol):
    """A protocol for objects that can be listed."""

    def list_all(self) -> list[str]:
        """Lists the items in the object."""
        ...


class ListableAndCountable(Countable, Listable):
    """A protocol for objects that can be counted and listed."""

    pass


class MatchboxDBAdapter(ABC):
    """An abstract base class for Matchbox database adapters."""

    settings: "MatchboxServerSettings"

    sources: ListableAndCountable
    models: Countable
    source_clusters: Countable
    model_clusters: Countable
    all_clusters: Countable
    creates: Countable
    merges: Countable
    proposes: Countable
    source_resolutions: Countable

    # Retrieval

    @abstractmethod
    def query(
        self,
        source: SourceResolutionPath,
        point_of_truth: ResolutionPath | None = None,
        threshold: int | None = None,
        return_leaf_id: bool = False,
        limit: int | None = None,
    ) -> Table:
        """Queries the database from an optional point of truth.

        Args:
            source: the resolution pathidentifying the source to query
            point_of_truth (optional): the resolution path to use for filtering results
                If not specified, will use the source resolution for the queried source
            threshold (optional): the threshold to use for creating clusters
                If None, uses the models' default threshold
                If an integer, uses that threshold for the specified model, and the
                model's cached thresholds for its ancestors
            return_leaf_id (optional): whether to return cluster ID of leaves
            limit (optional): the number to use in a limit clause. Useful for testing

        Returns:
            The resulting matchbox IDs in Arrow format
        """
        ...

    @abstractmethod
    def match(
        self,
        key: str,
        source: SourceResolutionPath,
        targets: list[SourceResolutionPath],
        point_of_truth: ResolutionPath,
        threshold: int | None = None,
    ) -> list[Match]:
        """Matches an ID in a source resolution and returns the keys in the targets.

        Args:
            key: The key to match from the source.
            source: The path of the source resolution.
            targets: The paths of the target source resolutions.
            point_of_truth: The path of the resolution to use for matching.
            threshold (optional): the threshold to use for creating clusters
                If None, uses the resolutions' default threshold
                If an integer, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors
                Will use these threshold values instead of the cached thresholds
        """
        ...

    # Collection management

    @abstractmethod
    def create_collection(self, name: CollectionName) -> Collection:
        """Create a new collection.

        Args:
            name: The name of the collection to create.

        Returns:
            A Collection object containing its metadata, versions, and resolutions.
        """

    @abstractmethod
    def get_collection(self, name: CollectionName) -> Collection:
        """Get collection metadata.

        Args:
            name: The name of the collection to get.

        Returns:
            A Collection object containing its metadata, versions, and resolutions.
        """
        ...

    @abstractmethod
    def list_collections(self) -> list[CollectionName]:
        """List all collection names.

        Returns:
            A list of collection names.
        """
        ...

    @abstractmethod
    def delete_collection(self, name: CollectionName, certain: bool) -> None:
        """Delete a collection and all its versions.

        Args:
            name: The name of the collection to delete.
            certain: Whether to delete the collection without confirmation.
        """
        ...

    # Version management

    @abstractmethod
    def create_run(self, collection: CollectionName) -> Run:
        """Create a new run.

        Args:
            collection: The name of the collection to create the run in.

        Returns:
            A Run object containing its metadata and resolutions.
        """
        ...

    @abstractmethod
    def set_run_mutable(
        self, collection: CollectionName, run_id: RunID, mutable: bool
    ) -> Run:
        """Set the mutability of a run.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run to update.
            mutable: Whether the run should be mutable.

        Returns:
            The updated Run object.
        """
        ...

    @abstractmethod
    def set_run_default(
        self, collection: CollectionName, run_id: RunID, default: bool
    ) -> Run:
        """Set the default status of a run.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run to update.
            default: Whether the run should be the default run.

        Returns:
            The updated Run object.
        """
        ...

    @abstractmethod
    def get_run(self, collection: CollectionName, run_id: RunID) -> Run:
        """Get run metadata and resolutions.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run to get.

        Returns:
            A Run object containing its metadata and resolutions.
        """
        ...

    @abstractmethod
    def delete_run(
        self, collection: CollectionName, run_id: RunID, certain: bool
    ) -> None:
        """Delete a run and all its resolutions.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run to delete.
            certain: Whether to delete the run without confirmation.
        """
        ...

    # Resolution management

    @abstractmethod
    def create_resolution(self, resolution: Resolution, path: ResolutionPath) -> None:
        """Writes a resolution to Matchbox.

        Args:
            resolution: Resolution object with a source or model config
            path: The resolution path for the source
        """
        ...

    @abstractmethod
    def get_resolution(self, path: ResolutionPath) -> Resolution:
        """Get a resolution from its path.

        Args:
            path: The resolution path for the source

        Returns:
            A Resolution object
        """
        ...

    @abstractmethod
    def update_resolution(self, resolution: Resolution, path: ResolutionPath) -> None:
        """Updates resolution metadata.

        It cannot be used to update a resolution's fingerprint.

        Args:
            resolution: Resolution object with a source or model config
            path: The resolution path for the source
        """
        ...

    @abstractmethod
    def delete_resolution(self, path: ResolutionPath, certain: bool) -> None:
        """Delete a resolution from the database.

        Args:
            path: The path of the resolution to delete.
            certain: Whether to delete the model without confirmation.
        """
        ...

    # Data insertion

    @abstractmethod
    def lock_resolution_data(self, path: ResolutionPath) -> None:
        """Change resolution upload stage to PROCESSING.

        This will lock uploading data.

        Args:
            path: The path of the resolution to target.
        """
        ...

    @abstractmethod
    def unlock_resolution_data(self, path: ResolutionPath, complete: bool) -> None:
        """Change resolution upload stage to READY.

        This will unlock uploading data.

        Args:
            path: The path of the resolution to target.
            complete: Whether to label the resolution stage as COMPLETE.
        """
        ...

    @abstractmethod
    def get_resolution_stage(self, path: ResolutionPath) -> UploadStage:
        """Retrieves upload stage of resolution data.

        Args:
            path: The path of the resolution to target.
        """
        ...

    @abstractmethod
    def insert_source_data(
        self, path: SourceResolutionPath, data_hashes: Table
    ) -> None:
        """Inserts hash data for a source resolution.

        Only possible if data fingerprint matches fingerprint declared when the
        resolution was created. Data can only be set once on a resolution.

        Args:
            path: The path of the source resolution to index.
            data_hashes: The Arrow table with the hash of each data row
        """
        ...

    @abstractmethod
    def insert_model_data(self, path: ModelResolutionPath, results: Table) -> None:
        """Inserts results data for a model resolution.

        Only possible if data fingerprint matches fingerprint declared when the
        resolution was created. Data can only be set once on a resolution.
        """
        ...

    @abstractmethod
    def get_model_data(self, path: ModelResolutionPath) -> Table:
        """Get the results for a model resolution."""
        ...

    # Data management

    @abstractmethod
    def validate_ids(self, ids: list[int]) -> bool:
        """Validates a list of IDs exist in the database.

        Args:
            ids: A list of IDs to validate.

        Raises:
            MatchboxDataNotFound: If some items don't exist in the target table.
        """
        ...

    @abstractmethod
    def dump(self) -> MatchboxSnapshot:
        """Dumps the entire database to a snapshot.

        Returns:
            A MatchboxSnapshot object of type "postgres" with the database's
                current state.
        """
        ...

    @abstractmethod
    def drop(self, certain: bool) -> None:
        """Hard clear the database by dropping all tables and re-creating.

        Args:
            certain: Whether to drop the database without confirmation.
        """
        ...

    @abstractmethod
    def clear(self, certain: bool) -> None:
        """Soft clear the database by deleting all rows but retaining tables.

        Args:
            certain: Whether to delete the database without confirmation.
        """
        ...

    @abstractmethod
    def restore(self, snapshot: MatchboxSnapshot) -> None:
        """Restores the database from a snapshot.

        Args:
            snapshot: A MatchboxSnapshot object of type "postgres" with the
                database's state

        Raises:
            TypeError: If the snapshot is not compatible with PostgreSQL
        """
        ...

    @abstractmethod
    def delete_orphans(self) -> int:
        """Deletes orphan clusters.

        Orphan clusters are clusters recorded in the Clusters table but that are
        not referenced in other tables.
        """
        ...

    # User, group and permissions management

    @abstractmethod
    def login(self, user: User) -> User:
        """Upserts the user to the database.

        Args:
            user: A User with a username and optionally an email
        """
        ...

    @abstractmethod
    def get_user_groups(self, user_name: str) -> list[GroupName]:
        """Get names of all groups a user belongs to.

        Args:
            user_name: The username to get the groups of
        """
        ...

    @abstractmethod
    def list_groups(self) -> list[Group]:
        """List all groups."""
        ...

    @abstractmethod
    def get_group(self, name: GroupName) -> Group:
        """Get group details including members.

        Args:
            name: The name of the group to fetch.

        Raises:
            MatchboxGroupNotFoundError: If group doesn't exist.
        """
        ...

    @abstractmethod
    def create_group(self, group: Group) -> None:
        """Create a new group.

        Arg:
            group: The group to create

        Raises:
            MatchboxGroupAlreadyExistsError: If group name taken.
        """
        ...

    @abstractmethod
    def delete_group(self, name: GroupName, certain: bool = False) -> None:
        """Delete a group.

        Args:
            name: The name of the group to delete
            certain: Whether to delete the group without confirmation.

        Raises:
            MatchboxGroupNotFoundError: If group doesn't exist.
            MatchboxSystemGroupError: If attempting to delete a system group.
            MatchboxDeletionNotConfirmed: If certain=False.
        """
        ...

    @abstractmethod
    def add_user_to_group(self, user_name: str, group_name: GroupName) -> None:
        """Add a user to a group. Creates user if they don't exist.

        Raises:
            MatchboxGroupNotFoundError: If group doesn't exist.
        """
        ...

    @abstractmethod
    def remove_user_from_group(self, user_name: str, group_name: GroupName) -> None:
        """Remove a user from a group.

        Raises:
            MatchboxGroupNotFoundError: If group doesn't exist.
            MatchboxUserNotFoundError: If user doesn't exist.
        """
        ...

    @abstractmethod
    def check_permission(
        self,
        user_name: str,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> bool:
        """Check user permission against a resource.

        Args:
            user_name: The username to check.
            permission: The permission type to check
            resource: The resource to check. One of "system" or a collection name
        """
        ...

    @abstractmethod
    def get_permissions(
        self,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> list[PermissionGrant]:
        """Get all granted permissions against a resource.

        Args:
            resource: The resource to get permissions for. One of "system" or a
                collection name
        """
        ...

    @abstractmethod
    def grant_permission(
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        """Grants a permission on a resource.

        Args:
            group_name: The name of the group to grant permission to
            permission: The permission to grant
            resource: The resource to grant permission on. One of "system" or a
                collection name
        """
        ...

    @abstractmethod
    def revoke_permission(
        self,
        group_name: GroupName,
        permission: PermissionType,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
    ) -> None:
        """Revoke permission on a resource.

        Args:
            group_name: The name of the group to revoke permission of
            permission: The permission to revoke
            resource: The resource to revoke permission from. One of "system" or a
                collection name
        """
        ...

    # Evaluation management

    @abstractmethod
    def insert_judgement(self, judgement: Judgement) -> None:
        """Adds an evaluation judgement to the database.

        Args:
            judgement: representation of the proposed clusters.
        """
        ...

    @abstractmethod
    def get_judgements(self, tag: str | None = None) -> tuple[Table, Table]:
        """Retrieves all evaluation judgements.

        Args:
            tag: optional string by which to filter judgements

        Returns:
            Two PyArrow tables with the judgments and their expansion.
            See `matchbox.common.arrow` for information on the schema.
        """
        ...

    @abstractmethod
    def sample_for_eval(
        self, n: int, path: ModelResolutionPath, user_name: str
    ) -> Table:
        """Sample a cluster to validate.

        Args:
            n: Number of clusters to sample
            path: Path of resolution from which to sample
            user_name: Name of user requesting the sample

        Returns:
            An Arrow table with the same schema as returned by `query()`
        """
        ...



"""Tests for authentication routes."""

from collections.abc import Callable
from typing import Any
from unittest.mock import Mock

import pytest
from fastapi.testclient import TestClient

from matchbox.client._settings import settings
from matchbox.common.dtos import (
    AuthStatusResponse,
    User,
)
from test.scripts.authorisation import generate_json_web_token


def test_login(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test the login endpoint at /auth/login."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.login = Mock(return_value=User(user_name="alice"))

    response = test_client.post(
        "/auth/login", json=User(user_name="alice").model_dump()
    )

    assert response.status_code == 200
    assert User.model_validate(response.json())


@pytest.mark.parametrize(
    ("has_token", "expected_authenticated"),
    [
        pytest.param(True, True, id="valid_token"),
        pytest.param(False, False, id="no_token"),
    ],
)
def test_auth_status(
    api_EdDSA_key_pair: tuple[bytes, bytes],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
    has_token: bool,
    expected_authenticated: bool,
) -> None:
    """Test auth status with and without valid JWT token."""
    test_client, _, mock_settings = api_client_and_mocks
    private_key, _ = api_EdDSA_key_pair
    mock_settings.authorisation = True

    username = "test.user@email.com"

    if has_token:
        token = generate_json_web_token(
            private_key_bytes=private_key,
            sub=username,
            api_root=settings.api_root,
        )
        test_client.headers["Authorization"] = token
    else:
        test_client.headers.pop("Authorization", None)

    response = test_client.get("/auth/status")

    assert response.status_code == 200
    result = AuthStatusResponse.model_validate(response.json())
    assert result.authenticated is expected_authenticated
    if expected_authenticated:
        assert result.username == username


PROTECTED_ROUTES = [
    pytest.param(
        "post",
        "/collections/default/runs/1/resolutions/name/data",
        id="post_resolution_data",
    ),
    pytest.param(
        "post", "/collections/default/runs/1/resolutions/name", id="post_resolution"
    ),
    pytest.param(
        "put", "/collections/default/runs/1/resolutions/name", id="put_resolution"
    ),
    pytest.param(
        "delete", "/collections/default/runs/1/resolutions/name", id="delete_resolution"
    ),
    pytest.param("delete", "/database", id="delete_database"),
]


@pytest.mark.parametrize(("method_name", "url"), PROTECTED_ROUTES)
def test_incorrect_signature(
    api_EdDSA_key_pair: tuple[bytes, bytes],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
    method_name: str,
    url: str,
) -> None:
    """Test that routes reject tokens with incorrect signatures."""
    test_client, _, _ = api_client_and_mocks
    private_key, _ = api_EdDSA_key_pair

    # Create token with wrong user but keep original signature
    _, _, signature_b64 = test_client.headers["Authorization"].encode().split(b".")
    header_b64, payload_64, _ = (
        generate_json_web_token(
            private_key_bytes=private_key,
            api_root=settings.api_root,
            sub="incorrect.user@email.com",
        )
        .encode()
        .split(b".")
    )
    test_client.headers["Authorization"] = b".".join(
        [header_b64, payload_64, signature_b64]
    ).decode()

    method: Callable[..., Any] = getattr(test_client, method_name)
    response = method(url)

    assert response.status_code == 401
    assert response.content == b'"JWT invalid."'


@pytest.mark.parametrize(("method_name", "url"), PROTECTED_ROUTES)
def test_expired_token(
    api_EdDSA_key_pair: tuple[bytes, bytes],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
    method_name: str,
    url: str,
) -> None:
    """Test that routes reject expired tokens."""
    test_client, _, _ = api_client_and_mocks
    private_key, _ = api_EdDSA_key_pair

    test_client.headers["Authorization"] = generate_json_web_token(
        private_key_bytes=private_key,
        sub="test.user@email.com",
        api_root=settings.api_root,
        expiry_hours=-2,
    )

    method: Callable[..., Any] = getattr(test_client, method_name)
    response = method(url)

    assert response.status_code == 401
    assert response.content == b'"JWT expired."'


@pytest.mark.parametrize(("method_name", "url"), PROTECTED_ROUTES)
def test_missing_authorisation_header(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
    method_name: str,
    url: str,
) -> None:
    """Test that routes reject requests without authorisation headers."""
    test_client, _, _ = api_client_and_mocks

    test_client.headers.pop("Authorization", None)

    method: Callable[..., Any] = getattr(test_client, method_name)
    response = method(url)

    assert response.status_code == 401
    assert response.content == b'"JWT required but not provided."'
